///////////////////////////////////////////////////////////////////////////////////////////////
//
//  This sample source code comes with Servosila SC-25C Brushless Motor Controllers.
//
//  The code is provided "AS IS" without any kind of guarantees or warranties.
//  Use it at your own risk.
//
//  The code is free for everyone to use, modify or redistribute.
//
//  www.servosila.com
//
///////////////////////////////////////////////////////////////////////////////////////////////

#ifndef SERVOSILA_SLCAN_DECODER_H
#define SERVOSILA_SLCAN_DECODER_H

#include <stddef.h> //size_t
#include <stdint.h>

// Examples:
//  t1230 : can_id 0x123, len 0, no data
//  t4563112233 : can_id 0x456, len 3, data 0x11 0x22 0x33
//  T12ABCDEF2AA55 : extended can_id 0x12ABCDEF, len 2, data 0xAA 0x55

namespace servosila
{

//Type definitions
//CAN payload - just an array of 8 bytes
typedef uint8_t payload_t[8];

//SLCAN commands
//  T iiiiiiii L dddddddddddddddd[CR]
class slcan_decoder
{
private:
    size_t          m_nsymbols;
    uint8_t         m_buffer[1 + 4*2 + 1 + 8*2 + 1];   //maximum size corresponds to a message with format "T iiiiiiii L dddddddddddddddd[CR]"
    bool            m_state;
    uint32_t        m_can_id;           //output
    payload_t       m_payload;          //output
    uint8_t         m_payload_length;   //output
public:
    //constructor
    slcan_decoder()
    {
        m_nsymbols  = 0;
        m_state     = false;
    }

    inline void reset()
    {
        m_nsymbols  = 0;
        m_state     = false;
    }

    inline void push_or_reset(uint8_t symbol)
    {
        if(m_nsymbols<sizeof(m_buffer))     //buffer overflow control
        {
            //push the symbol into the buffer
            m_buffer[m_nsymbols] = symbol;
            m_nsymbols++;
        }
        else
        {   //buffer overflow while waiting for [CR]
            reset();
        }
    }

    //The function returns TRUE if a complete and valid SLCAN 'T'- or 't'-type message has been received.
    inline bool process_symbol(uint8_t symbol)
    {
        bool result = false;

        if((symbol=='t') ||         //11bit identifier or
           (symbol=='T'))           //29bit identifier
        {
            //restarting & adding to the buffer
            m_buffer[0] = symbol;
            m_nsymbols  = 1;
            m_state     = true;     //state "waiting for [CR]"
        }
        else if(((symbol>='0') && (symbol<='9')) || ((symbol>='A') && (symbol<='F')))
        {
            if(m_state)
            {
                //ADDING A KNOWN SYMBOL INTO THE BUFFER
                push_or_reset(symbol);
            }
            else
            {
                //A STRAY KNOWN SYMBOL
                reset();
            }
        }
        else if((symbol==13) || (symbol==10)) // [CR]
        {
            if(m_state)
            {
                //A MESSAGE CANDIDATE RECEIVED SUCCESFULLY
                result = _parse();
                //Reset to wait for the next command
                reset();
                //ATTENTION: CR is not added to the buffer; keep this in mind when parsing the buffer
            }
            else
            {
                //A RANDOM [CR] RECEIVED
                reset();
            }
        }
        else
        {
            //UNKNOWN SYMBOL
            reset();
        }

        return result;
    }

    //getters
    const payload_t& get_payload() const {   return m_payload;      };
          uint32_t   get_can_id()  const {   return m_can_id;       };

private:
    inline bool _parse()
    {
        bool result = false;

        if     (m_buffer[0] == 't') result = _parse_11bit();    //11bit identifier
        else if(m_buffer[0] == 'T') result = _parse_29bit();    //29bit identifier

        return result;
    }

    inline bool _parse_11bit()
    {
        bool result = false;

        if(m_nsymbols>=5)       //the minimal message is t 001 0 [CR]  (CR is not added to the buffer)
        {
            //Decoding CAN-ID
            m_can_id  = hex2uint32(m_buffer[3]);
            m_can_id |= hex2uint32(m_buffer[2]) << 4ul*1ul;
            m_can_id |= hex2uint32(m_buffer[1]) << 4ul*2ul;

            //Decoding Payload Length
            const size_t len = hex2uint32(m_buffer[4]);
            m_payload_length = len;

            //Decoding the Payload
            if((len<=8) && (1+3+1+2*len==m_nsymbols))   //checking if payload length is correct
            {
                if(len>=1) m_payload[0] = hex2uint32(m_buffer[ 5]) << 4ul | hex2uint32(m_buffer[ 6]); else goto out;
                if(len>=2) m_payload[1] = hex2uint32(m_buffer[ 7]) << 4ul | hex2uint32(m_buffer[ 8]); else goto out;
                if(len>=3) m_payload[2] = hex2uint32(m_buffer[ 9]) << 4ul | hex2uint32(m_buffer[10]); else goto out;
                if(len>=4) m_payload[3] = hex2uint32(m_buffer[11]) << 4ul | hex2uint32(m_buffer[12]); else goto out;
                if(len>=5) m_payload[4] = hex2uint32(m_buffer[13]) << 4ul | hex2uint32(m_buffer[14]); else goto out;
                if(len>=6) m_payload[5] = hex2uint32(m_buffer[15]) << 4ul | hex2uint32(m_buffer[16]); else goto out;
                if(len>=7) m_payload[6] = hex2uint32(m_buffer[17]) << 4ul | hex2uint32(m_buffer[18]); else goto out;
                if(len>=8) m_payload[7] = hex2uint32(m_buffer[19]) << 4ul | hex2uint32(m_buffer[20]); else goto out;
                out:
                result = true;
            }
        }

        return result;
    }

    inline bool _parse_29bit()
    {
        bool result = false;

        if(m_nsymbols>=10)      //the minimal message is T 12345678 0 [CR]  (CR is not added to the buffer)
        {
            //Decoding CAN-ID
            m_can_id  = hex2uint32(m_buffer[8]);
            m_can_id |= hex2uint32(m_buffer[7]) << 4ul*1ul;
            m_can_id |= hex2uint32(m_buffer[6]) << 4ul*2ul;
            m_can_id |= hex2uint32(m_buffer[5]) << 4ul*3ul;
            m_can_id |= hex2uint32(m_buffer[4]) << 4ul*4ul;
            m_can_id |= hex2uint32(m_buffer[3]) << 4ul*5ul;
            m_can_id |= hex2uint32(m_buffer[2]) << 4ul*6ul;
            m_can_id |= hex2uint32(m_buffer[1]) << 4ul*7ul;

            //Decoding Payload Length
            const size_t len = hex2uint32(m_buffer[9]);
            m_payload_length = len;

            //Decoding the Payload
            if((len<=8) && (1+8+1+2*len==m_nsymbols))   //checking if payload length is correct
            {
                if(len>=1) m_payload[0] = hex2uint32(m_buffer[10]) << 4ul | hex2uint32(m_buffer[11]); else goto out;
                if(len>=2) m_payload[1] = hex2uint32(m_buffer[12]) << 4ul | hex2uint32(m_buffer[13]); else goto out;
                if(len>=3) m_payload[2] = hex2uint32(m_buffer[14]) << 4ul | hex2uint32(m_buffer[15]); else goto out;
                if(len>=4) m_payload[3] = hex2uint32(m_buffer[16]) << 4ul | hex2uint32(m_buffer[17]); else goto out;
                if(len>=5) m_payload[4] = hex2uint32(m_buffer[18]) << 4ul | hex2uint32(m_buffer[19]); else goto out;
                if(len>=6) m_payload[5] = hex2uint32(m_buffer[20]) << 4ul | hex2uint32(m_buffer[21]); else goto out;
                if(len>=7) m_payload[6] = hex2uint32(m_buffer[22]) << 4ul | hex2uint32(m_buffer[23]); else goto out;
                if(len>=8) m_payload[7] = hex2uint32(m_buffer[24]) << 4ul | hex2uint32(m_buffer[25]); else goto out;
                out:
                result = true;
            }
        }

        return result;
    }

    inline uint32_t hex2uint32(uint8_t symbol) const
    {
        uint32_t result = 0;

        if((symbol>='0') && (symbol<='9'))
        {
            result = symbol - '0';
        }
        else if((symbol>='A') && (symbol<='F'))
        {
            result = symbol - 'A' + 10;
        }
        else
        {
            //ERROR: the control flow should never come here
            //assert(false);
        }

        return result;
    }

}; //class

} //namespace

#endif
